/*
     This program synchronizes the simple producer-consumer
     example in the program 1-trivialPC-unsynchronized.c.

     What we have is a producer-consumer example with a bounded
     buffer of size one (not a very big buffer). This program uses
     two binary semaphores to do the synchronization by "bounding"
     the buffer.

     With the semaphores, the "producer" cannot increment the
     counter until the "consumer" has printed out the current
     value of the counter. And the "consumer" cannot print out
     another value until the "producer" has done another increment.
     So the output of this program is all of the integers in
     consecutive order.

     Notice that the problem in 1-trivialPC-unsynchronized.c is
     not a critical section problem. We cannot solve the problem in
     1-trivialPC-unsynchronized.c by using a mutex. The synchronization
     done below with the semaphores is a different synchronization
     pattern than the "mutual exclusion" needed for critical sections.
*/
#include <windows.h>   // Win32 Sleep()
#include <unistd.h>    // Linux sleep()
#include <stdio.h>
#include "pthread.h"
#include "semaphore.h"

// global variables
static int runFlag = TRUE;
static int count = 0;

sem_t semaphoreI;  // the incremented semaphore
sem_t semaphoreP;  // the printed semaphore

int main( int argc, char *argv[])
{
   //thread function prototypes
   void *thread1(void *vargp);
   void *thread2(void *vargp);

   pthread_t thread1ID;
   pthread_t thread2ID;
   int threadParam1 = 0;  // thread function parameter
   int threadParam2 = 0;

   // create two Semaphore objects (one signaled, one not signaled)
   sem_init(&semaphoreP, 0, 1); // signaled
   sem_init(&semaphoreI, 0, 0); // not signaled

   // create threads
   printf("Starting threads\n");
   pthread_create(&thread1ID, NULL, thread1, &threadParam1);  // should have error checking
   pthread_create(&thread2ID, NULL, thread2, &threadParam2);

   printf("The incrementer is thread number %d.\n", thread1ID);
   printf("The printer is thread number %d.\n", thread2ID);

   Sleep(800);     // let the threads run for a while
   runFlag = FALSE; // now stop the threads
   Sleep(200);      // give them time to stop
   printf("\nThreads stopped.\n");
   printf("count = %d\n", count);

   return 0;
}//main


void *thread1(void *vargp)
{
   while(runFlag)
   {
      sem_wait(&semaphoreP);
      count++;
      sem_post(&semaphoreI);
   }
   return NULL;
}//thread1


void *thread2(void *vargp)
{
   while(runFlag)
   {
      sem_wait(&semaphoreI);
      printf("%d ", count);
      sem_post(&semaphoreP);
   }
   return NULL;
}//thread2
