/*
    This program uses one counting semaphores to bound the consumer (the
    counting semaphore is in addition to the binary semaphore used to
    protect the critical section).

    Each time the consumer wants to remove an item, it has to first check
    semaphoreFull to see if there is at least one item filling up a space. After
    the consumer has removed an item, it increments semaphoreEmpty to signal
    that there is another empty space.





    This program starts out with a half full buffer. How would you modify it to
    start with an empty buffer? With a full buffer?
*/
#include <windows.h>   // Win32 Sleep()
#include <unistd.h>    // Linux sleep()
#include <stdio.h>
#include "pthread.h"
#include "semaphore.h"

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];

sem_t semaphoreCS;     // critical section semaphore
sem_t semaphoreFull;   // bounding semaphore


int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   void *putIn(void *vargp);
   void *takeOut(void *vargp);

   pthread_t producerID;
   pthread_t consumerID;
   int threadParam1 = 0;  // thread function parameter
   int threadParam2 = 0;

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create a binary Semaphore to protect the critical section
   sem_init(&semaphoreCS, 0, 1); // signaled

   // create a counting Semaphore that signals a half full buffer
   sem_init(&semaphoreFull,  0, 39); // signaled


   // create the producer and consumer threads
   printf("Starting producer and consumer threads\n");
   pthread_create(&producerID, NULL, putIn,   &threadParam1);  // should have error checking
   pthread_create(&consumerID, NULL, takeOut, &threadParam2);

   printf("The producer is thread number %d.\n", producerID);
   printf("The consumer is thread number %d.\n", consumerID);

   Sleep(50000);     // let the threads run for a while
   runFlag = FALSE;  // now stop the threads
   Sleep(200);       // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
void *putIn(void *vargp)
{
   int r; // will hold a random integer
   // seed the random number generator for this thread
   srand( GetTickCount() );

   while(runFlag)
   {



     // begin a critical section
      sem_wait(&semaphoreCS);
      theBuffer[bufferIndex+1] = '+';
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      sem_post(&semaphoreCS);
     // end of critical section

      // Signal that an item has been placed in the buffer
      sem_post(&semaphoreFull);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(70);  // sleep a short while
   }
   return 0;
}//putIn


// remove a plus sign from the buffer (the consumer)
void *takeOut(void *vargp)
{
   int r;  // will hold a random integer
   // seed the random number generator for this thread
   srand( 2*GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an item in the buffer to remove
      sem_wait(&semaphoreFull);

     // begin a critical section
      sem_wait(&semaphoreCS);
      theBuffer[bufferIndex] = '0';
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      sem_post(&semaphoreCS);
     // end of critical section




      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(70);  // sleep a short while
   }
   return 0;
}//takeOut