/*
     This program uses a binary semaphore to protect the critical section
     in the producer-consumer problem.

     The buffer does not get out of whack right away as in the last example.
     But if this program runs long enough, the producer will overflow the buffer
     and then the buffer gets out of whack again. We need a way to bound the
     producer so that it does not put too many items in the buffer (and we need
     to bound the consumer so that it does not try to remove too many items).
*/
#include <windows.h>
#include <stdio.h>

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];

HANDLE semaphoreCS;  // critical section semaphore


int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   DWORD WINAPI putIn(LPVOID);
   DWORD WINAPI takeOut(LPVOID);

   DWORD producerID;
   DWORD consumerID;
   int producerParam = 0;  // thread function parameters
   int consumerParam = 0;

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create a binary Semaphore to protect the critical section
   semaphoreCS = CreateSemaphore(NULL, 1, 1, NULL); // signaled

   // create the producer and consumer threads
   printf("Starting producer and consumer threads\n");
   CreateThread(NULL, 0, putIn, &producerParam, 0, &producerID);
   CreateThread(NULL, 0, takeOut, &consumerParam, 0, &consumerID);

   printf("The producer is thread number %d.\n", producerID);
   printf("The consumer is thread number %d.\n", consumerID);

   Sleep(5000);     // let the threads run for a while
   runFlag = FALSE; // now stop the threads
   Sleep(200);      // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
DWORD WINAPI putIn(LPVOID threadParam)
{
   while(runFlag)
   {
     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex+1] = '+';
      Sleep(0); // give up the CPU
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      Sleep(70);  // sleep a short while
      //Sleep(20);  // or sleep a shorter while
   }
   return 0;
}//putIn


// remove a plus sign from the buffer (the consumer)
DWORD WINAPI takeOut(LPVOID threadParam)
{
   while(runFlag)
   {
     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex] = '0';
      Sleep(0); // give up the CPU
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      Sleep(120);  // sleep a short while
   }
   return 0;
}//takeOut