/*
    This program uses two counting semaphores to bound the producer and
    consumer (the counting semaphores are in addition to the binary semaphore
    used to protect the critical section).

    Each time the producer wants to produce an item, it has to first check
    semaphoreEmpty to see if there is an empty space. After the producer produces
    an item it increments semaphoreFull to signal that a space has been filled.

    Each time the consumer wants to remove an item, it has to first check
    semaphoreFull to see if there is at least one item filling up a space. After
    the consumer has removed an item, it increments semaphoreEmpty to signal
    that there is another empty space.

    This program starts out with a half full buffer. How would you modify it to
    start with an empty buffer? With a full buffer?
*/
#include <windows.h>
#include <stdio.h>

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];

HANDLE semaphoreCS;     // critical section semaphore
HANDLE semaphoreEmpty;  // bounding semaphores
HANDLE semaphoreFull;


int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   DWORD WINAPI putIn(LPVOID);
   DWORD WINAPI takeOut(LPVOID);

   DWORD producerID;
   DWORD consumerID;
   int producerParam = 0;  // thread function parameters
   int consumerParam = 0;

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create a binary Semaphore to protect the critical section
   semaphoreCS = CreateSemaphore(NULL, 1, 1, NULL); // signaled

   // create two counting Semaphores that signal a half full buffer
   semaphoreEmpty = CreateSemaphore(NULL, 38, 77, NULL); // signaled
   semaphoreFull = CreateSemaphore(NULL, 39, 77, NULL);  // signaled

   // create the producer and consumer threads
   printf("Starting producer and consumer threads\n");
   CreateThread(NULL, 0, putIn, &producerParam, 0, &producerID);
   CreateThread(NULL, 0, takeOut, &consumerParam, 0, &consumerID);

   printf("The producer is thread number %d.\n", producerID);
   printf("The consumer is thread number %d.\n", consumerID);

   Sleep(50000);     // let the threads run for a while
   runFlag = FALSE;  // now stop the threads
   Sleep(200);       // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
DWORD WINAPI putIn(LPVOID threadParam)
{
   int r; // will hold a random integer
   // seed the random number generator for this thread
   srand( GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an empty spot in the buffer
      WaitForSingleObject(semaphoreEmpty, INFINITE);

     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex+1] = '+';
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      // Signal that an item has been placed in the buffer
      ReleaseSemaphore(semaphoreFull, 1, NULL);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(70);  // sleep a short while
   }
   return 0;
}//putIn


// remove a plus sign from the buffer (the consumer)
DWORD WINAPI takeOut(LPVOID threadParam)
{
   int r;  // will hold a random integer
   // seed the random number generator for this thread
   srand( 2*GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an item in the buffer to remove
      WaitForSingleObject(semaphoreFull, INFINITE);

     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex] = '0';
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      // Signal an empty spot in the buffer
      ReleaseSemaphore(semaphoreEmpty, 1, NULL);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(120);  // sleep a short while
   }
   return 0;
}//takeOut