/*
    This program implements a producer-consumer type of synchronization where the producers and consumers alternate filling and then emptying the buffer. That is, the producers fill the buffer while the consumers all block. When the buffer is full, the producers all block, the consumers wake up, and the producers remain blocked until the consumers empty the buffer.
*/
#include <windows.h>
#include <stdio.h>

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];

HANDLE semaphoreCS;     // critical section semaphore
HANDLE semaphoreEmpty;  // bounding semaphores
HANDLE semaphoreFull;

int fullSlots = 39;    // counts the number of buffer slots that are full
                       // the buffer will have 77 slots
                       // the buffer will be initialized to be half full

int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   DWORD WINAPI putIn(LPVOID);
   DWORD WINAPI takeOut(LPVOID);

   DWORD producerID1;
   DWORD consumerID1;
   DWORD producerID2;
   DWORD consumerID2;
   char producerParam1 = '+';  // thread function parameters
   char consumerParam1 = '0';
   char producerParam2 = 'x';  // thread function parameters
   char consumerParam2 = '@';

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create a binary Semaphore to protect the critical section
   semaphoreCS = CreateSemaphore(NULL, 1, 1, NULL); // signaled

   // create two counting Semaphores that signal a half full buffer
   semaphoreEmpty = CreateSemaphore(NULL, 0, 77, NULL);   // not signaled
   semaphoreFull =  CreateSemaphore(NULL, 39, 77, NULL);  // signaled

   // create the producer and consumer threads
   printf("Starting two producer and two consumer threads\n");
   CreateThread(NULL, 0, putIn,   &producerParam1, 0, &producerID1);
   CreateThread(NULL, 0, takeOut, &consumerParam1, 0, &consumerID1);
   CreateThread(NULL, 0, putIn,   &producerParam2, 0, &producerID2);
   CreateThread(NULL, 0, takeOut, &consumerParam2, 0, &consumerID2);

   printf("The producer is thread number %d.\n", producerID1);
   printf("The consumer is thread number %d.\n", consumerID1);
   printf("The producer is thread number %d.\n", producerID2);
   printf("The consumer is thread number %d.\n", consumerID2);

   Sleep(50000);     // let the threads run for a while
   runFlag = FALSE;  // now stop the threads
   Sleep(200);       // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
DWORD WINAPI putIn(LPVOID threadParam)
{
   int r; // will hold a random integer
   // seed the random number generator for this thread
   srand( GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an empty spot in the buffer
      WaitForSingleObject(semaphoreEmpty, INFINITE);

     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex+1] = *((char*)threadParam);
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      fullSlots++;
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      // Signal the consumers if the buffer is full
      if (fullSlots == 77)  // the buffer is full
         ReleaseSemaphore(semaphoreFull, 77, NULL);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(70);  // sleep a short while
   }
   return 0;
}//putIn


// remove a plus sign from the buffer (the consumer)
DWORD WINAPI takeOut(LPVOID threadParam)
{
   int r;  // will hold a random integer
   // seed the random number generator for this thread
   srand( 2*GetTickCount() );

   while(runFlag)
   {
      // Make sure there is an item in the buffer to remove
      WaitForSingleObject(semaphoreFull, INFINITE);

     // begin a critical section
      WaitForSingleObject(semaphoreCS, INFINITE);
      theBuffer[bufferIndex] = *((char*)threadParam);
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
      fullSlots--;
      ReleaseSemaphore(semaphoreCS, 1, NULL);
     // end of critical section

      // Signal producers if the buffer is empty
      if (fullSlots == 0)  // the buffer is empty
         ReleaseSemaphore(semaphoreEmpty, 77, NULL);

      // sleep a random amount
      r = rand()%4;
      if (r == 0) Sleep(10*(rand()%10)^10);

      //Sleep(120);  // sleep a short while
   }
   return 0;
}//takeOut